/// <reference types="webpack-env" />
import MessageType, { MessageTypeData, IMessageType } from './MessageType';
/**
 * Setup events auto-loading for all instances of this class
 *
 * The class needs to inherit MessageEmitter to take effect
 *
 * @param module Node.js/Webpack module
 * @param folder Path relative to this module to the folder containing events files
 */
export declare function AutoloadEvents(require: __WebpackModuleApi.RequireContext): (target: any) => void;
/**
 * Configuration options for message event emitters
 */
export interface IEmitterConfig {
    send?: (buffer: ArrayBuffer) => void;
}
/**
 * List of events other than messages that users may listen on
 */
export declare const enum Event {
    Error = "error",
    Ignored = "ignored",
}
/**
 * Callback for Event.Error events
 */
export declare type ErrorEventCallback = (error: Error) => any;
/**
 * Event callback
 */
export declare type EventCallback<T> = (message: T) => any;
/**
 * MessageEmitter class
 *
 * MessageEmitter instances can be used to simplify listening to
 * and emitting messages.
 */
export default class MessageEmitter {
    /**
     * Known events
     *
     * List of events we have either listened to or
     */
    protected knownEvents: {
        [eventName: string]: boolean;
    };
    /**
     * Events loaded using the auto-loading mechanism
     *
     * @see `@AutoloadEvents`
     */
    protected autoloadedEvents: {
        [eventName: string]: <T extends MessageEmitter>(instance: T) => void;
    };
    /**
     * Internal function
     */
    private _send?;
    /**
     * Internal event emitter
     */
    private emitter;
    /**
     * Require instance to use during auto-loading
     */
    private readonly require;
    constructor(config?: IEmitterConfig);
    /**
     * Message Event listener
     *
     * @param type
     * @param callback
     */
    on(type: Event.Error, callback: ErrorEventCallback): void;
    on(type: Event.Ignored, callback: EventCallback<any>): void;
    on<T extends MessageType>(type: IMessageType<T>, callback: EventCallback<T>): void;
    /**
     * Message Event listener
     *
     * @param type
     * @param callback
     */
    once(type: Event.Error, callback: ErrorEventCallback): void;
    once(type: Event.Ignored, callback: EventCallback<any>): void;
    once<T extends MessageType>(type: IMessageType<T>, callback: EventCallback<T>): void;
    /**
     * Synchronously emit a message
     *
     * @param message
     */
    emit<T extends MessageType>(message: T): boolean;
    /**
     * Asynchronously emit a message
     *
     * @param message
     */
    emitAsync<T extends MessageType>(message: T): Promise<any[]>;
    /**
     * Utility to create a message parser
     */
    createMessageParser(): (buffer: ArrayBuffer) => Promise<void>;
    /**
     * Send message using the configured send function
     *
     * @param type
     * @param data
     */
    send<T extends MessageType>(type: IMessageType<T>, data: MessageTypeData<T>): void;
    /**
     * Prepare a message emission
     *
     * This method will attempt event file auto-loading, and
     * emit an Ignored event if the message type has never
     * been listened to by this listener
     *
     * @param message
     */
    private prepareEmit<T>(message);
    /**
     * Emit message as ignored
     *
     * @param message
     */
    private emitIgnored<T>(message);
    /**
     * Attempt to automatically load event listeners from a file
     *
     * @param name
     */
    private autoload(name);
    /**
     * Process auto-load errors
     *
     * @param name
     */
    private handleAutoloadError(name, error);
}
