/// <reference types="webpack-env" />
/**
 * Type attributes contain the information
 */
export declare type TypeAttribute = [string, string];
/**
 * Interface representing the static methods and attributes
 * of the MessageType class
 */
export interface IMessageType<T> {
    name: string;
    id: number;
    size: number;
    attributes: TypeAttribute[];
    pool: any[];
    new (): T;
    getTypeClassById<T extends MessageType>(_id: number): IMessageType<T> | undefined;
    claim<T extends MessageType>(): T;
    release<T extends MessageType>(instance: T): void;
    create<T extends MessageType>(this: IMessageType<T>, data?: any): T;
    parse<T extends MessageType>(buffer: ArrayBuffer): T;
}
export declare type Diff<T extends string, U extends string> = ({
    [P in T]: P;
} & {
    [P in U]: never;
} & {
    [x: string]: never;
})[T];
export declare type Omit<T, K extends keyof T> = Pick<T, Diff<keyof T, K>>;
/**
 * Message Type Data Type
 *
 * This is a type representing an instance of the message type without
 * any inherited methods or attributes attached to it.
 */
export declare type MessageTypeData<T extends MessageType> = Omit<T, keyof MessageType> & {
    [x: string]: any;
};
/**
 * Message type base class
 *
 * This class serves two purposes
 *
 *   1. Serialization/deserialization
 *   2. Resource management (object and buffer pooling)
 */
export default class MessageType {
    /**
     * Reference to the user defined TypeIds enum
     */
    static typeIds: {
        [id: number]: string;
    };
    /**
     * ID to type classes map
     */
    static typesRegister: Map<number, typeof MessageType>;
    /**
     * Unique identifier when messages of this class are packed
     *
     * See `../types:TypeIds`
     */
    static readonly id: number;
    /**
     * Total size in bytes of unpacked messages of this class
     */
    static readonly size: number;
    /**
     * Transmission type map
     *
     * This is used at runtime to decide how a given
     * value will be transported. For instance, numbers
     * can be boiled down to Uint8 when sent over the wire
     * so to reduce the overall data size.
     */
    static readonly attributes: TypeAttribute[];
    /**
     * Pools
     */
    static pool: any[];
    /**
     * Require instance used during auto-loading (if present)
     */
    static require?: __WebpackModuleApi.RequireContext;
    /**
     * Dataview instance
     *
     * This will contain serialization/deserialization data
     */
    private _data?;
    /**
     * Retrieve the class for a type using the type ID
     *
     * @param id
     */
    static getTypeClassById<T extends MessageType>(id: number): IMessageType<T> | undefined;
    /**
     * Claim an instance of this class (or create one if none are available).
     */
    static claim<T extends MessageType>(): T;
    /**
     * Release an instance back into the object pool
     *
     * @param instance
     */
    static release<T extends MessageType>(instance: T): void;
    /**
     * Create an instance of this class, and apply data to it
     *
     * @param data
     */
    static create<T extends MessageType>(this: IMessageType<T>, data?: MessageTypeData<T>): T;
    /**
     * Parse a given buffer and deserialize a message
     *
     * @param buffer
     */
    static parse<T extends MessageType>(buffer: ArrayBuffer): T;
    constructor();
    /**
     * Unpack a given dataview into an object of this class instance
     *
     * @param view DataView
     */
    unpack(view: DataView): void;
    /**
     * Pack this object instance into a DataView for transmission
     */
    pack(): ArrayBuffer;
    /**
     * Release this object instance back into the object pool.
     */
    release(): void;
    /**
     * Internal function
     *
     * This will be overwritten at runtime with code generated from
     * the information cumulated in `MessageType.attributes`.
     */
    private _create<T>(_instance);
    /**
     * Internal function
     *
     * This will be overwritten at runtime with code generated from
     * the information cumulated in `MessageType.attributes`.
     *
     * @param view DataView
     * @param obj
     */
    private _unpack(_instance, _data);
    /**
     * Internal function
     *
     * This will be overwritten at runtime with code generated from
     * the information cumulated in `MessageType.attributes`.
     *
     * @param view DataView
     */
    private _pack(_instance, _data);
}
